# -*- coding: utf-8 -*-
"""
Created on Wed Sep 19 13:06:35 2018

@author: hrajiv
"""
import pandas as pd
import sys
import numpy as np
import avarana

name='ts_forecast'

def lag(array, start_index, end_index, lag):
    lag_start_index = start_index-lag
    lag_end_index = end_index-lag
    yyy = array[lag_start_index:lag_end_index]
    return yyy

def mean_absolute_percentage_error(y_true, y_pred): 
    y_true[y_true==0] = y_true[y_true>0].min() 
    h1 = np.abs(y_true - y_pred)
    h2 = np.divide(h1, y_true)
    return np.mean(h2)

def tf_mape(y_true, y_pred):
    y_true[y_true==0] = y_true[y_true>0].min() 
    h1 = np.abs(y_true - y_pred)
    h2 = np.divide(h1, y_true)
    return np.mean(h2)

def weighted_absolute_percentage_error(y_true, y_pred): 
    return np.abs(y_true - y_pred).sum()/y_true.sum()

def prediction_interval(y, y_pred):
    
    y = y.astype(float).copy()
    y = y.astype(float).copy()
    
    if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
        y = y.values
        
    if isinstance(y_pred, pd.DataFrame) or isinstance(y_pred, pd.Series):
        y_pred = y_pred.values
    
    # estimate stdev of yhat
    sum_errs = np.sum((y - y_pred)**2)
    stdev = np.sqrt(1/(len(y)-2) * sum_errs)

    # calculate prediction interval
    interval = 1.96 * stdev
   
    return interval

def test_forecast_ARNN(data, lags=[], verbose=0):
      
    # Used for error messages
    signature = avarana.get_function_name()

    if data.count() == 0:
        return -1
    
    first_index = data.first_valid_index()    
    data.drop(index=range(0,first_index), axis=0, inplace=True)
    data.reset_index(inplace=True, drop=True)
    
    if len(lags) == 0:
        print(signature, ': Lags cannot be empty', file=sys.stderr)
        return None, -1
    
    n = max(lags)
    data_len = len(data)
    observed_len = data.last_valid_index() + 1
    
    # Set training lenght to twice the number of periods in year to allow for lag
    train_len = 2*n
    
    if observed_len < train_len:
        print(signature, ': Not enough training observations. Need a minimum of %i observations' %train_len, file=sys.stderr)
        return None, -1 
    
    if data.count() < 0.5*observed_len:
        print(signature, ': Minimum of half of the training values should be unique' %train_len, file=sys.stderr)
        return -3    
    
    # Based on the last non null value
    test_len = data.last_valid_index() - train_len + 1 

    # If there are no observations for forecast 
    if data_len <= train_len:
        print(signature, ': No data in forecast period', file=sys.stderr)
        return None, -1
     
    # Replace missing with zeros
    data.fillna(0, inplace=True)

    target = data.astype(float).copy()        

    lag_columns = ['X%s' % x for x in lags]
    X_lag_df = pd.DataFrame(columns = lag_columns)
    
    for i in range(len(lags)):
        lag_temp = lag(target.copy(), n, train_len, lags[i])
        X_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)

    X = X_lag_df.copy()
    y = target[n:train_len].copy()
    
    X_test_lag_df = pd.DataFrame(columns = lag_columns)    
    for i in range(len(lags)):
        lag_temp = lag(target.copy(), train_len, train_len + test_len, lags[i])
        X_test_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)

    X_test = X_test_lag_df.copy()    
    y_test = target[train_len:train_len+test_len].reset_index(drop=True).copy()    
    
    # Model configuration
    mape = -1
    model, predictions, mape = avarana.nnet_regressor(X, y,
                                                      X_test, y_test, 
                                                      hidden_layers=[n],
                                                      activation='relu',
                                                      loss='mape',
                                                      verbose=verbose)
    
    for j in range(train_len, data_len):
        X_fcst_lag_df = pd.DataFrame(columns = lag_columns)    
        for i in range(len(lags)):
            lag_temp = lag(target.copy(), j, j+1, lags[i])
            X_fcst_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)        

        X_fcst = X_fcst_lag_df.copy()
        
        Y_fcst = model.predict(X_fcst.values)
        target[j] = Y_fcst[0][0]
    
    mape = avarana.mean_absolute_percentage_error(y_test.values, target[train_len:observed_len].values)
    return target[train_len:data_len].copy(), mape

def test_forecast_ARNNEXOG(data, target_col, lags=[], verbose=0):
    
    # Used for error messages
    signature = avarana.get_function_name()
    
    if target_col is None:
        print(signature, ': Target variable is undefined', file=sys.stderr)
        return None, -1
    
    if data[target_col].count() == 0:
        print(signature, ': Target variable does not have sufficient data', file=sys.stderr)        
        return None, -1
    
    if target_col not in data.columns:
        print(signature, ': ', target_col, ' not found in data', file=sys.stderr)
        return None,-1
    
    if len(lags) == 0:
        print(signature, ': Lags cannot be empty to ', signature, file=sys.stderr)
        return None, -1
    
    first_index = data[target_col].first_valid_index()    
    data.drop(index=range(0,first_index), axis=0, inplace=True)
    data.reset_index(inplace=True, drop=True)	
	
    n = max(lags)
    
    data_len = len(data)
    observed_len = data[target_col].last_valid_index() + 1
    
    # Twice the number of periods in year to allow for lag
    train_len = 2*n
    
    if observed_len < train_len:
        print(signature, 'Not enough training observations. Need a minimum of %i observations' %(train_len), file=sys.stderr)
        return None,-1 
    
    # Based on the last non null value
    test_len = data[target_col].last_valid_index() - train_len + 1 

    # If there are no observations for forecast 
    if data_len <= train_len:
        print(signature, ': No data in forecast period', file=sys.stderr)
        return None,-1
    
    # Replace missing with zeros
    data.fillna(0, inplace=True)

    target = data[target_col].astype(float).copy()        
    data.drop(target_col, axis=1, inplace=True)
        
    lag_columns = ['X%s' % x for x in lags]
    X_lag_df = pd.DataFrame(columns = lag_columns)
    
    for i in range(len(lags)):
        lag_temp = lag(target.copy(), n, train_len, lags[i])
        X_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)

    X = data[n:train_len].reset_index(drop=True).copy()
    Z = pd.concat([X_lag_df,X], axis=1, sort=False)
    X = Z  
    y = target[n:train_len].copy()
    
    X_test_lag_df = pd.DataFrame(columns = lag_columns)    
    for i in range(len(lags)):
        lag_temp = lag(target.copy(), train_len, train_len + test_len, lags[i])
        X_test_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)

    X_test = data[train_len:train_len+test_len].reset_index(drop=True).copy()
    Z = pd.concat([X_test_lag_df, X_test], axis=1, sort=False)
    X_test = Z
    
    y_test = target[train_len:train_len+test_len].reset_index(drop=True).copy()    
    
    # Model configuration
    mape = -1
    model, predictions, mape = avarana.nnet_regressor(X, y,
                                                      X_test, y_test, 
                                                      hidden_layers=[n],
                                                      activation='relu',
                                                      loss='mape',
                                                      verbose=verbose)
    for j in range(train_len, data_len):
        X_fcst_lag_df = pd.DataFrame(columns = lag_columns)    
        for i in range(len(lags)):
            lag_temp = lag(target.copy(), j, j+1, lags[i])
            X_fcst_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)        

        X_fcst=data[j:j+1].reset_index(drop=True).copy()
        Z = pd.concat([X_fcst_lag_df,X_fcst], axis=1, sort=False)
        X_fcst = Z
        
        Y_fcst = model.predict(X_fcst.values)
        target[j] = Y_fcst[0][0]
    
    
    return target[train_len:data_len].copy(), mape    

def forecast_ARNNEXOG(data, target_col, lags=[], verbose=0):
    
    if target_col is None:
        print('Target variable is undefined', file=sys.stderr)
        return None, -1
    
    if target_col not in data.columns:
        print(target_col, ' not found in data', file=sys.stderr)
        return None,-1
    
    # Used for filenames
    signature = avarana.get_function_name()
    if len(lags) == 0:
        print('Lags cannot be empty to ', signature, file=sys.stderr)
        return None, -1
    
    n = max(lags)
    data_len = len(data)
    observed_len = data[target_col].last_valid_index() + 1
    
    train_len = observed_len
    
    if observed_len < n:
        print('Not enough training observations. Need a minimum of %i observations' %(train_len), file=sys.stderr)
        return None,-1 
    
    # Replace missing with zeros
    data.fillna(0, inplace=True)

    target = data[target_col].astype(float).copy()        
    data.drop(target_col, axis=1, inplace=True)
        
    lag_columns = ['X%s' % x for x in lags]
    X_lag_df = pd.DataFrame(columns = lag_columns)
    
    for i in range(len(lags)):
        lag_temp = lag(target.copy(), n, train_len, lags[i])
        X_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)

    X = data[n:train_len].reset_index(drop=True).copy()
    Z = pd.concat([X_lag_df,X], axis=1, sort=False)
    X = Z  
    y = target[n:train_len].copy()
    
    # Set test to null so no mape is calculated
    X_test = []    
    y_test = []
    
    # Model configuration
    mape = -1
    model, predictions, mape = avarana.nnet_regressor(X, y,
                                                      X_test, y_test, 
                                                      hidden_layers=[n],
                                                      activation='relu',
                                                      loss='mape',
                                                      verbose=verbose)
    # Calculate forecast values
    for j in range(train_len, data_len):
        X_fcst_lag_df = pd.DataFrame(columns = lag_columns)    
        for i in range(len(lags)):
            lag_temp = lag(target.copy(), j, j+1, lags[i])
            X_fcst_lag_df[lag_columns[i]] = lag_temp.reset_index(drop=True)        

        X_fcst=data[j:j+1].reset_index(drop=True).copy()
        Z = pd.concat([X_fcst_lag_df,X_fcst], axis=1, sort=False)
        X_fcst = Z
        
        Y_fcst = model.predict(X_fcst.values)
        target[j] = Y_fcst[0][0]
    
    
    return target[train_len:data_len].copy()    


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 19 13:06:35 2018

@author: hrajiv
"""
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Dropout
from keras.layers import Embedding
from keras.layers import LSTM
from keras import optimizers
import pandas as pd
import sys
import numpy as np
import xgboost
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.metrics import explained_variance_score
from eli5.permutation_importance import get_score_importances
from sklearn.model_selection import GridSearchCV
import traceback

name='avarana'

def get_function_name():
    return traceback.extract_stack(None, 2)[0][2]

from tensorflow import set_random_seed
set_random_seed(7)

from tensorflow import set_random_seed
set_random_seed(2)

def parser(x):
	return pd.datetime.strptime(x, '%Y%m%d')

def lag(array, start_index, end_index, lag):
    lag_start_index = start_index-lag
    lag_end_index = end_index-lag
    yyy = array[lag_start_index:lag_end_index]
    return yyy

def mean_absolute_percentage_error(y_true, y_pred): 
    y_true[y_true==0] = y_true[y_true>0].min() 
    h1 = np.abs(y_true - y_pred)
    h2 = np.divide(h1, y_true)
    return np.mean(h2)

def tf_mape(y_true, y_pred):
    y_true[y_true==0] = y_true[y_true>0].min() 
    h1 = np.abs(y_true - y_pred)
    h2 = np.divide(h1, y_true)
    return np.mean(h2)

def weighted_absolute_percentage_error(y_true, y_pred): 
    return np.abs(y_true - y_pred).sum()/y_true.sum()

def nnet_regressor(X, y, X_test=[], y_test=[], hidden_layers=[],
                   activation='relu', loss='mape', verbose=0):
    if type(hidden_layers) is not list:
        print("Hidden layers must be a list")
        return None
        
    model = Sequential()
    
    if len(hidden_layers) > 0 and np.sum(hidden_layers) > 0:
        first_layer = 0
        for layer_size in hidden_layers:
            if layer_size > 0 and first_layer == 0:
                model.add(Dense(layer_size, input_dim=X.shape[1], activation='linear'))
                first_layer = 1
            elif layer_size > 0:
                model.add(Dense(layer_size, activation='relu'))
        model.add(Dense(1, activation='linear')) #hard coded to 1 output layer
    else:
        model.add(Dense(1, input_dim=X.shape[1], activation='linear')) #hard coded to 1 output layer

    opt_fun = 'rmsprop'
    opt_fun = optimizers.Nadam(lr=0.002) 
    model.compile(loss=loss, optimizer=opt_fun)

    def scorer(X, y):
        y_pred = model.predict(X)
        return 1-mean_absolute_percentage_error(y, y_pred)    
    
#    gsc = GridSearchCV(
#        estimator=model,
#        param_grid={
#            #'neurons': range(18,31,4),
#            'noise': [x/20.0 for x in range(3, 7)],
#        },
#        scoring='neg_absolute_error',
#        cv=5
#    )   
    
    
    
#    last_min_mape = 100
#    last_min_layer = []
#    
#    hidden_layers = []
#    for h in range(0,X.shape[1]):
#        hidden_layers.append(0)
#        for i in range(0,len(X)+1): 
#            hidden_layers[h] = i
#            model, predictions, mape = manthan_nnet(X,Y,X_test,test,hidden_layers)
#            if mape < last_min_mape:
#                last_min_layer = hidden_layers
#                last_min_mape = mape
#    
#    print('Min layer size = ',last_min_layer, ' last min MAPE= ', last_min_mape)
    
    model.fit(X.values, y.values, batch_size=round(len(y)/2), epochs=50, verbose=verbose)
       
#   Set defaults if there is no test period in the inputs
    predictions = []
    mape = 0
    
    if len(X_test) > 0:
       if verbose > 0:
           base_score, score_decreases = get_score_importances(scorer, X_test.values, y_test.values)
           feature_importances = np.mean(score_decreases, axis=0)
           abs_feature_importances = np.abs(feature_importances )
           feature_importances_table = pd.DataFrame(np.transpose([X.columns, feature_importances, abs_feature_importances]), 
                                                    columns=['Feature','Score', 'Abs Score'])
           print(feature_importances_table.sort_values('Abs Score', ascending=False)) 
        
       predictions = model.predict(X_test).ravel()
       error = mean_squared_error(y_test.values, predictions)
       mape = mean_absolute_percentage_error(y_test.values, predictions)
       wape = weighted_absolute_percentage_error(y_test.values, predictions)
       r2 = r2_score(y_test.values, predictions)
       if verbose > 0:
           print('Hidden layers: ',hidden_layers,' R2: %.3f Test MSE: %.3f ~MAPE: %.6f WAPE: %.3f' % (r2, error, mape, wape))    
    
    return model, predictions, mape;

# define base model
def baseline_model():
	# create model
	model = Sequential()
	model.add(Dense(5, input_dim=5, kernel_initializer='normal', activation='relu'))
	model.add(Dense(1, kernel_initializer='normal'))
	# Compile model
	model.compile(loss='mean_squared_error', optimizer='adam')
	return model

# XGB Configuration
def xgb_regressor(X, Y, X_test, Y_test):
    xgb = xgboost.XGBRegressor(n_estimators=100, 
                           learning_rate=0.1, 
                           gamma=0.1, 
                           subsample=0.8,
                           colsample_bytree=0.75, 
                           max_depth=3,
                           min_child_weight=1)

    xgb.fit(X.values,Y.values)
    
    param_test1 = {
     'max_depth':range(3,100,2),
     'min_child_weight':range(1,6,2)
#     'learning_rate':np.arange(0.1,1,0.1),
#     'gamma':[i/10.0 for i in range(0,5)],
#     'subsample':[i/10.0 for i in range(6,10)],
#     'colsample_bytree':[i/10.0 for i in range(6,10)]
     }

    gsearch1 = GridSearchCV(estimator = xgb, 
                        param_grid = param_test1, 
                        n_jobs=1,iid=False, 
                        cv=5)

    gsearch1.fit(X.values,Y.values)

    print('Grid Score')
    print(gsearch1.grid_scores_)
    print('Best params')
    print(gsearch1.best_params_)
    print('Best Score')
    print(gsearch1.best_score_)
    
    predictions = xgb.predict(X_test.values)
    mape = mean_absolute_percentage_error(Y_test.values, predictions)
    evs = explained_variance_score(predictions, Y_test.values)
    return xgb, predictions, mape, evs

def lstm_regressor(X, Y, X_test, Y_test):
    max_features = 1024
    
    model = Sequential()
    model.add(Embedding(max_features, output_dim=256))
    model.add(LSTM(128))
    model.add(Dropout(0.5))
    model.add(Dense(1, activation='sigmoid'))
    
    model.compile(loss='mape',
                  optimizer='rmsprop',
                  metrics=['accuracy'])
    
    model.fit(x_train, y_train, batch_size=16, epochs=10)
    score = model.evaluate(x_test, y_test, batch_size=16)
